; Copyright (c) 2010, James Bailie <jimmy@mammothcheese.ca>.
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
;
;     * Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following disclaimer.
;     * Redistributions in binary form must reproduce the above copyright
; notice, this list of conditions and the following disclaimer in the
; documentation and/or other materials provided with the distribution.
;     * The name of James Bailie may not be used to endorse or promote
; products derived from this software without specific prior written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
; LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.

; -----------------------------------------------------------------------

; responder.mm -- skeletal plunger(8) responder.  This program is a
; modified version of cgi.mm.  Plunger is a generic SCGI server which
; allows ordinary filter programs to act as SCGI application servers.
; Plunger is distributed with the drood(8) web server.

; -------------------------------------------------------------------------

#include "stacks.mm"

(setq nl (char 10))
(setq terminator (concat "__END__" nl))
(setq scgi_body "")

; Contains SCGI environment.

(setq scgi_env (table))

; Contains decoded SCGI parameters.

(setq scgi_data (table))

; Contains cookies.

(setq scgi_cookies (table))

; scgi_encode - encodes a string in x-www-form-urlencoding

(let ((reserved (concat "]$;:@&+!?=#{}/[^`~\"<>|\\" (char 9) (char 10) (char 13)))

      (r2 (regcomp " "))
      (r3 (regcomp "%"))
      (r4 (regcomp "\\\\"))

      (adjust (lambda (d)
                 (cond ((eq d "10") "A")
                       ((eq d "11") "B")
                       ((eq d "12") "C")
                       ((eq d "13") "D")
                       ((eq d "14") "E")
                       ((eq d "15") "F")
                       (1 d))))

      (regexps (table)))

   (let ((rx (regcomp (concat "[" (substitute r4 "\\\\\\\\" reserved 1) "]"))))

      (apply_stack
         (lambda (x)
            (let ((item (stack)))
               (push item (regcomp (if (or (eq x ">") (eq x "<")) x (concat "\\" x))))
               (push item (concat "%" (adjust (stringify (/ (code x) 16)))
                                      (adjust (stringify (% (code x) 16)))))
               (hash regexps x item)))

         (explode reserved))

      (setq scgi_encode_helper
         (lambda (m x)
            (if (not m)
               x

               (letn ((r (lookup regexps (index m 0)))
                      (s (substitute (index r 0) (index r 1) x 0)))

                  (scgi_encode_helper (matches rx s) s)))))

      (setq scgi_encode
         (lambda (x)
            (let ((s (substitute r3 "%25" x 0)))
               (substitute r2 "+" (scgi_encode_helper (matches rx s) s) 0))))))

; scgi_decode - decodes a string encoded in x-www-form-urlencoding.

(let ((regexps (table))
      (r0 (regcomp "%([0-9A-Fa-f][0-9A-Fa-f])"))
      (r1 (regcomp "\\+"))
      (r2 (regcomp ".")))

   (setq scgi_decode
      (lambda (x)
         (scgi_decode_helper (substitute r1 " " x 0) (match r0 x) (matches r0 x))))

   (setq scgi_decode_helper
      (lambda (x indices encoding)
         (if (not encoding)
            x

            (let ((items (lookup regexps (index encoding 0))))

               (if items
                  (concat (substitute (index items 0)
                                      (index items 1)
                                      (substring x 0 (index indices 1))
                                      1)
                          (if (eq (length x) (index indices 1))
                             ""
                             (let ((x (substring x (index indices 1) 0)))
                                (scgi_decode_helper x (match r0 x) (matches r0 x)))))

                  (letn ((rx (regcomp (index encoding 0)))

                         (hex_digit_to_decimal
                            (lambda (x)
                               (let ((c (code (substitute r2 "\u\&" x))))
                                  (if (and (>= c 65) (<= c 70))
                                     (- c 55)
                                     (digitize x)))))

                         (digits (map_stack hex_digit_to_decimal (explode (index encoding 1))))
                         (converted (char (+ (* 16 (index digits 0)) (index digits 1))))
                         (new (stack)))

                     (push new rx)
                     (push new converted)
                     (hash regexps (index encoding 0) new)
                     (scgi_decode_helper x indices encoding))))))))

; Composes a complete Set-Cookie header.
; pairs is a stack of key/value pair strings.
; path is a string
; ttl can be a string or a number.

(setq scgi_make_cookie
   (lambda (pairs path ttl)
      (let ((cookie (scgi_make_cookie_helper pairs "")))
         (concat "Set-Cookie: " cookie "Version=1; Path=" path "; Max-Age=" (stringify ttl)))))

(setq scgi_make_cookie_helper
   (lambda (pairs text)
      (if (not (used pairs))
         text
         (scgi_make_cookie_helper pairs (concat text (shift pairs) "=" (shift pairs) "; ")))))

; Initializes scgi_cookies.

(let ((leading_rx (regcomp "^[\b\t]+(.*)[\b\t]+$")))

   (setq scgi_get_cookies
      (lambda ()

         ; Clear out any old cookies from the last invocation.

         (apply_stack (lambda (x) (unhash scgi_cookies x)) (keys scgi_cookies))

         (let ((cookie (lookup scgi_env "HTTP_COOKIE")))

            (when cookie

               (apply_stack
                  (lambda (x)
                     (hash scgi_cookies (substitute leading_rx "\1" (index x 0))
                                        (substitute leading_rx "\1" (index x 1))))

                  (map_stack
                     (lambda (x) (split "=" x))
                     (split ",;" cookie))))))))

; Sends the HTTP and HTML headers to the client.

(let ((nl (char 10)) (cr (char 13)))

   (setq scgi_send_headers
      (lambda (title stylesheet script cookie favicon extras)

         (println "Content-Type: text/html" cr)

         (when cookie
            (println cookie cr))

         (when extras
            (apply_stack (lambda (x) (println x cr)) extras))

         ; Header separator.

         (print cr nl)

         (println "<!DOCTYPE html" nl
            "   PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"" nl
            "   \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">" nl

            "<html xmlns=\"http://www.w3.org/1999/xhtml\">" nl

            "<head>" nl
            "<title>" title "</title>")

         (when stylesheet
            (println "<link rel=\"stylesheet\" type=\"text/css\" media=\"screen\" href=\"" stylesheet "\" />"))

         (when script
            (println "<script type=\"text/javascript\" src=\"" script "\"></script>"))

         (when favicon
            (println "<link rel=\"shortcut icon\" href=\"" favicon "\" />"))

         (println "</head>" nl
            "<body>"))))

; Initializes scgi_data.

(setq scgi_get_data
   (lambda ()

      ; Clear out any old parameters from the last invocation.

      (apply_stack (lambda (x) (unhash scgi_data x)) (keys scgi_data))

      ; Get the new parameters.

      (let ((d (if (eq (lookup scgi_env "REQUEST_METHOD") "GET")
                   (lookup scgi_env "QUERY_STRING")
                   scgi_body)))

         (when d
            (apply_stack
               (lambda (x)
                  (hash scgi_data
                     (scgi_decode (index x 0))
                     (if (eq (used x) 2) (scgi_decode (index x 1)) "")))

               (map_stack
                  (lambda (x) (split "=" x))
                  (split "&" d)))))))

; Sends the contents of a table as an HTML table.

(setq scgi_send_items
   (lambda (k h)
      (when (used k)
         (println "<tr><td>" (index k 0) ":</td><td>"
            (lookup h (shift k)) "</td></tr>")
         (scgi_send_items k h))))

; Invoked once for each client request.

(setq scgi_service_request
   (lambda ()
      (scgi_send_headers "Minimunger Test Responder" "" "" "" "" "")

      (println "<h1>Minimunger Test Responder</h1>" nl
         "<h2>SCGI Environment</h2>" nl "<table>")
      (scgi_send_items (keys scgi_env) scgi_env)

      (println "</table>" nl "<h2>SCGI Parameters</h2>" nl "<table>")
      (scgi_send_items (keys scgi_data) scgi_data)

      (println "</table>" nl "<h2>Cookies</h2>" nl "<table>")
      (scgi_send_items (keys scgi_cookies) scgi_cookies)
      
      (println "</table>" nl "</body>" nl "</html>" nl "__END__")
      (flush)))

; Reads request bodies.

(setq scgi_read_body
   (lambda (line)
      (if (and line (not (eq line terminator)))
         (progn
            (setq scgi_body (concat scgi_body line))
            (scgi_read_body (getline)))
         
         (setq scgi_body (chop scgi_body))
         (scgi_get_data)
         (scgi_get_cookies)
         (scgi_service_request))))
   
; Main loop.

(setq scgi_process_requests
   (lambda (line)
      (when line

         (let ((line (chomp line)))
            
            (if line

               ; We don't clear out the old scgi_env values, because
               ; the web server sends the same set with every request.
               ; The old are overwritten by the new.
               
               (letn ((pair (split ":" line 2))
                      (var (index pair 0))
                      (val (index pair 1)))
   
                  (hash scgi_env var val))

               (setq scgi_body "")
               (scgi_read_body (getline))))

         (scgi_process_requests (getline)))))

(scgi_process_requests (getline))
